import matplotlib as mpl
import numpy as np
import pandas as pd
from lxml import etree

mpl.use("Agg")
import contextlib

import matplotlib.pyplot as plt
from matplotlib import rcParams

rcParams.update({"figure.autolayout": True})
plt.style.use("bmh")


def smear_data(data, sigma):
    """
    Apply Gaussian smearing to spectrum y value.

    Args:
        sigma: Std dev for Gaussian smear function
    """
    from scipy.ndimage.filters import gaussian_filter1d

    diff = [data[i + 1, 0] - data[i, 0] for i in range(len(data) - 1)]
    avg_x_per_step = np.sum(diff) / len(diff)
    data[:, 1] = gaussian_filter1d(data[:, 1], sigma / avg_x_per_step)
    return data


class units:
    am2kg = 1.6605402e-27
    ev2j = 1.60217733e-19
    plank = 6.626075e-34
    c = 2.99792458e10
    pi = 3.1415926
    proton_mass = 1836
    ev2hartree = 27.211386245988
    a2bohr = 0.529
    ev2cm = 8065.6


class vasprun:
    """
    parse vasprun.xml and return all useful info to self.values

    Args:
        vasp_file: the path of vasprun.xml
        verbosity: output error msgs or not
    """

    def __init__(self, vasp_file="vasprun.xml", verbosity=0):
        self.error = False
        self.errormsg = ""
        self.values = {}
        try:
            doc = etree.parse(vasp_file)
            doc = doc.getroot()
            self.parse_vaspxml(doc)
            self.get_band_gap()
            N_atom = len(self.values["finalpos"]["positions"])
            self.values["calculation"]["energy_per_atom"] = self.values["calculation"]["energy"] / N_atom
        except etree.XMLSyntaxError:
            self.error = True
            self.errormsg = "corrupted file found"

        if verbosity > 0 and self.error is True:
            print("----------Warning---------------")
            print(self.errormsg)
            print("--------------------------------")

    def parse_vaspxml(self, xml_object):
        """
        parse the following tags

        - incar
        - kpoints
        - atominfo - composition, elements, formula
        - calculation - eigenvalues, energy, dos, fermi energy, stress, force
        - finalpos - lattice, rec_lat, positions
        """

        for child in xml_object.iterchildren():
            self.values.setdefault(child.tag, {})
            if child.tag == "incar":
                self.values[child.tag] = self.parse_i_tag_collection(child)
            elif child.tag == "kpoints":
                self.values[child.tag] = self.parse_kpoints(child)
            elif child.tag == "parameters":
                self.values[child.tag] = self.parse_parameters(child)
            elif child.tag == "atominfo":
                self.values["name_array"] = self.parse_name_array(child)
                self.values["composition"] = self.parse_composition(child)
                self.values["elements"] = self.get_element(self.values["composition"])
                self.values["formula"] = self.get_formula(self.values["composition"])
                (
                    self.values["pseudo_potential"],
                    self.values["potcar_symbols"],
                    self.values["valence"],
                    self.values["mass"],
                ) = self.get_potcar(child)
            elif child.tag == "calculation":
                self.values["calculation"], scf_count = self.parse_calculation(child)
                if self.values["parameters"]["electronic"]["electronic convergence"]["NELM"] == scf_count:
                    self.error = True
                    self.errormsg = "SCF is not converged"

                if (
                    self.values["parameters"]["electronic"]["electronic spin"]["LSORBIT"]
                    or self.values["parameters"]["electronic"]["electronic spin"]["ISPIN"] == 2
                ):
                    self.spin = True
                else:
                    self.spin = False
            elif child.tag == "structure" and child.attrib.get("name") == "finalpos":
                self.values["finalpos"] = self.parse_finalpos(child)
            elif child.tag not in (
                "i",
                "r",
                "v",
                "incar",
                "kpoints",
                "atominfo",
                "calculation",
            ):
                self.values[child.tag] = self.parse_vaspxml(child)
            # else:
            #    return 1
            self.dict_clean(self.values)

    @staticmethod
    def dict_clean(dict_del):
        """
        Delete the keys that is {} and None
        Loops recursively over nested dictionaries.
        """
        dict_foo = dict_del.copy()  # Used as iterator to avoid the 'DictionaryHasChanged' error
        for key in dict_foo:
            if isinstance(dict_foo[key], dict):
                vasprun.dict_clean(dict_del[key])

            # if len(dict_foo[key])==0:
            if dict_foo[key] == {} or dict_foo[key] is None:
                with contextlib.suppress(KeyError):
                    del dict_del[key]

        return dict_del

    def parse_finalpos(self, finalpos):
        """
        obtain final configuration
        """
        d = {}
        for i in finalpos.iter("varray"):
            name = i.attrib.get("name")
            d[name] = self.parse_varray(i)
        return d

    def parse_dynmat(self, dynmat):
        hessian, eigenvalues, eigenvectors = [], [], []
        for va in dynmat.findall("varray"):
            if va.attrib.get("name") == "hessian":
                hessian = self.parse_varray(va)
            elif va.attrib.get("name") == "eigenvectors":
                eigenvectors = self.parse_varray(va)
        factor = np.sqrt(units.ev2j / 1e-20 / units.am2kg)
        for v in dynmat.findall("v"):
            for i in v.text.split():
                eigenvalues.append(np.sqrt(abs(float(i))) * factor * units.plank / units.ev2j / 2 / units.pi)
                # eigenvalues.append(np.sqrt(abs(float(i))))
        return hessian, eigenvalues, eigenvectors

    def parse_born_chg(self, charge):
        """
        obtain the born charge
        """
        chg = []
        for info in charge.findall("set"):
            chg.append(self.parse_varray(info))
        return chg

    def parse_i_tag_collection(self, itags_collection):
        d = {}
        for info in itags_collection.findall("i"):
            name = info.attrib.get("name")
            type = info.attrib.get("type")
            content = info.text
            d[name] = self.assign_type(type, content)
        return d

    @staticmethod
    def parse_varray_pymatgen(elem):
        def _vasprun_float(f):
            """
            Large numbers are often represented as ********* in the vasprun.
            This function parses these values as np.nan
            """
            try:
                return float(f)
            except ValueError as e:
                f = f.strip()
                if f == "*" * len(f):
                    warnings.warn("Float overflow (*******) encountered in vasprun")
                    return np.nan
                raise e

        if elem.get("type", None) == "logical":
            m = [[i == "T" for i in v.text.split()] for v in elem]
        else:
            m = [[_vasprun_float(i) for i in v.text.split()] for v in elem]

        return m

    @staticmethod
    def parse_varray(varray):
        if varray.get("type") == "int":
            m = [[int(number) for number in v.text.split()] for v in varray.findall("v")]
        else:
            try:
                m = [[float(number) for number in v.text.split()] for v in varray.findall("v")]
            except ValueError:
                m = [[0 for number in v.text.split()] for v in varray.findall("v")]
        return m

    @staticmethod
    def parse_array(array):
        array_dictionary = {}
        values = []
        dimension_list = {}
        field_list = []

        for dimension in array.findall("dimension"):
            dimension_list[dimension.attrib.get("dim")] = dimension.text

        for field in array.findall("field"):
            field_list.append(field.text)

        for r in array.iter("r"):
            values.append([float(number) for number in r.text.split()])

        array_dictionary["value"] = values
        array_dictionary["dimensions"] = dimension_list
        array_dictionary["fields"] = field_list

        return array_dictionary

    @staticmethod
    def assign_type(type, content):
        if type == "logical":
            content = content.replace(" ", "")
            if content in ("T", "True", "true"):
                return True
            elif content in ("F", "False", "false"):
                return False
            else:
                Warning("logical text " + content + " not T, True, true, F, False, false, set to False")
            return False
        elif type == "int":
            return int(content) if len(content.split()) == 1 else [int(number) for number in content.split()]
        elif type == "string":
            return content
        elif type is None:
            return float(content) if len(content.split()) == 1 else [float(number) for number in content.split()]
        else:
            Warning("New type: " + type + ", set to string")
        return content

    @staticmethod
    def parse_composition(atom_info):
        atom_names = {}
        for set in atom_info.findall("array"):
            if set.attrib.get("name") == "atoms":
                for rc in set.iter("rc"):
                    atom_name = rc.find("c").text.replace(" ", "")
                    if atom_name in atom_names:
                        atom_names[atom_name] += 1
                    else:
                        atom_names[atom_name] = 1

                break
        return atom_names

    @staticmethod
    def get_element(atom_names_dictionary):
        elements = []
        for atom_name in atom_names_dictionary:
            elements.append(atom_name.replace(" ", ""))
        return elements

    @staticmethod
    def get_formula(atom_names_dictionary):
        formula = ""
        for atom_name in atom_names_dictionary:
            formula += atom_name.replace(" ", "") + str(atom_names_dictionary[atom_name])
        return formula

    def get_potcar(self, child):
        """
        parse the potcar information

        - {'labels': ['O', 'Sr_sv'], 'pot_type': 'paw', 'functional': 'pbe'}
        - ['PAW_PBE', 'N', '08Apr2002']
        """
        pseudo = {"labels": [], "pot_type": [], "functional": []}
        potcar_symbol = []
        valence = []
        mass = []
        for i in child.iterchildren():
            if i.tag == "array" and i.attrib.get("name") == "atomtypes":
                ll = list(i.iter("c"))
                for i in range(3, len(ll), 5):
                    valence.append(float(ll[i].text))

                for i in range(2, len(ll), 5):
                    mass.append(float(ll[i].text))

                for i in range(4, len(ll), 5):
                    text = ll[i].text.split()
                    label = text[1]
                    pot = text[0].split("_")[0]
                    try:
                        xc = text[0].split("_")[1]
                    except:
                        xc = "unknown"
                    pseudo["labels"].append(label)
                    pseudo["pot_type"].append(pot)
                    pseudo["functional"].append(xc)
                    potcar_symbol.append(xc + " " + label)
        return pseudo, potcar_symbol, valence, mass

    @staticmethod
    def parse_name_array(atominfo):
        atom_names = []
        for array in atominfo.findall("array"):
            if array.attrib["name"] == "atoms":
                atom_names = [rc.find("c").text.strip() for rc in array.find("set")]

        if atom_names == []:
            ValueError("No atomname found in file")

        return atom_names

    #    def parse_eigenvalue(self, eigenvalue):
    #        eigenvalue = eigenvalue.find("array")
    #        eigenvalues = self.parse_array(eigenvalue)
    #        return eigenvalues
    def parse_parameters(self, child):
        parameters = {}
        for i in child:
            if i.tag == "separator":
                name = i.attrib.get("name")
                d = self.parse_i_tag_collection(i)
                parameters[name] = d
                for ii in i:
                    if ii.tag == "separator":
                        name2 = ii.attrib.get("name")
                        d2 = self.parse_i_tag_collection(ii)
                        parameters[name][name2] = d2
        return parameters

    def parse_eigenvalue(self, eigenvalue):
        eigenvalues = []
        for s in eigenvalue.find("array").find("set").findall("set"):
            for ss in s.findall("set"):
                eigenvalues.append(self.parse_varray_pymatgen(ss))
        return eigenvalues

    def parse_dos(self, dos):
        t_dos = []
        p_dos = []
        for s in dos.find("total").find("array").findall("set"):
            for ss in s.findall("set"):
                t_dos.append(self.parse_varray_pymatgen(ss))
        if dos.find("partial") is not None and len(dos.find("partial")) > 0:
            for s in dos.find("partial").find("array").findall("set"):
                for _i, ss in enumerate(s.findall("set")):
                    p = []
                    for sss in ss.findall("set"):
                        p.append(self.parse_varray_pymatgen(sss))
                    p_dos.append(p)

        return t_dos, p_dos

    def parse_projected(self, proj):
        projected = []
        for s in proj.find("array").find("set").findall("set"):
            for ss in s.findall("set"):
                p = []
                for sss in ss.findall("set"):
                    p.append(self.parse_varray_pymatgen(sss))
                projected.append(p)

        return projected  # [N_kpts, N_bands]

    def parse_calculation(self, calculation):
        stress = []
        force = []
        efermi = 0.0
        eigenvalues = []
        energy = 0.0
        scf_count = 0
        tdos = []
        pdos = []
        born_chgs = []
        hessian = []
        dyn_eigenvalues = []
        dyn_eigenvectors = []
        epsilon_ion = []
        proj = []
        for i in calculation.iterchildren():
            if i.attrib.get("name") == "stress":
                stress = self.parse_varray(i)
            elif i.attrib.get("name") == "forces":
                force = self.parse_varray(i)
            elif i.tag == "dos":
                for j in i.findall("i"):
                    if j.attrib.get("name") == "efermi":
                        efermi = float(j.text)
                        break
                tdos, pdos = self.parse_dos(i)
            elif i.tag == "projected":
                proj = self.parse_projected(i)
            elif i.tag == "eigenvalues":
                eigenvalues = self.parse_eigenvalue(i)
            elif i.tag == "scstep":
                for j in i.iterchildren():
                    if j.tag == "energy":
                        for e in j.findall("i"):
                            if e.attrib.get("name") == "e_fr_energy":
                                scf_count += 1

            elif i.tag == "energy":
                for e in i.findall("i"):
                    if e.attrib.get("name") == "e_fr_energy":
                        try:
                            energy = float(e.text)
                        except ValueError:
                            energy = 100000000
                    else:
                        Warning("No e_fr_energy found in <calculation><energy> tag, energy set to 0.0")
            elif i.tag == "array" and i.attrib.get("name") == "born_charges":
                born_chgs = self.parse_born_chg(i)
            elif i.tag == "varray" and i.attrib.get("name") == "epsilon_ion":
                epsilon_ion = self.parse_varray(i)

            elif i.tag == "dynmat":
                hessian, dyn_eigenvalues, dyn_eigenvectors = self.parse_dynmat(i)

        calculation = {}
        calculation["stress"] = stress
        calculation["efermi"] = efermi
        calculation["force"] = force
        calculation["eband_eigenvalues"] = eigenvalues
        calculation["energy"] = energy
        calculation["tdos"] = tdos
        calculation["pdos"] = pdos
        calculation["projected"] = proj
        calculation["born_charges"] = born_chgs
        calculation["hessian"] = hessian
        calculation["normal_modes_eigenvalues"] = dyn_eigenvalues
        calculation["normal_modes_eigenvectors"] = dyn_eigenvectors
        calculation["epsilon_ion"] = epsilon_ion
        return calculation, scf_count

    def parse_kpoints(self, kpoints):
        kpoints_dict = {"list": [], "weights": [], "divisions": [], "mesh_scheme": ""}

        for i in kpoints.iterchildren():
            if i.tag == "generation":
                kpoints_dict["mesh_scheme"] = i.attrib.get("param")
                for j in i.iterchildren():
                    if j.attrib.get("name") == "divisions":
                        kpoints_dict["divisions"] = [int(number) for number in j.text.split()]
                        break

        for va in kpoints.findall("varray"):
            name = va.attrib["name"]
            if name == "kpointlist":
                kpoints_dict["list"] = self.parse_varray(va)
            elif name == "weights":
                kpoints_dict["weights"] = self.parse_varray(va)

        return kpoints_dict

    def get_bands(self):
        """
        Function for computing the valence band index from the count of electrons

        Args:
            None

        Returns:
            bands: an integer number
            occupy: bool number
        """
        self.values["valence"]
        self.values["composition"]
        total = int(self.values["parameters"]["electronic"]["NELECT"])

        # if self.spin:
        #    fac = 1
        # else:
        #    fac = 2
        fac = 2

        if total % 2 == 0:
            IBAND = int(total / fac)
            occupy = True
        else:
            IBAND = int(total / fac) + 1
            occupy = False

        self.values["bands"] = IBAND
        self.values["occupy"] = occupy

    @staticmethod
    def get_cbm(kpoints, efermi, eigens, IBAND):
        ind = np.argmin(eigens[:, IBAND, 0])
        pos = kpoints[ind]
        value = eigens[ind, IBAND, 0] - efermi
        return {"kpoint": pos, "value": value}

    @staticmethod
    def get_vbm(kpoints, efermi, eigens, IBAND):
        ind = np.argmax(eigens[:, IBAND, 0])
        pos = kpoints[ind]
        value = eigens[ind, IBAND, 0] - efermi
        return {"kpoint": pos, "value": value}

    def get_band_gap(self):
        self.get_bands()
        IBAND = self.values["bands"]
        occupy = self.values["occupy"]
        self.values["metal"] = False
        self.values["gap"] = None
        self.values["cbm"] = None
        self.values["vbm"] = None
        if occupy is True:
            efermi = self.values["calculation"]["efermi"]
            eigens = np.array(self.values["calculation"]["eband_eigenvalues"])
            kpoints = np.array(self.values["kpoints"]["list"])
            if np.shape(eigens)[0] > np.shape(kpoints)[0]:
                kpoints = np.tile(kpoints, [2, 1])

            cbm = self.get_cbm(kpoints, efermi, eigens, IBAND)
            vbm = self.get_vbm(kpoints, efermi, eigens, IBAND - 1)
            self.values["gap"] = cbm["value"] - vbm["value"]
            self.values["cbm"] = cbm
            self.values["vbm"] = vbm
            if self.values["gap"] < 0:
                self.values["metal"] = True
                self.values["gap"] = 0
        else:
            self.values["metal"] = True
            self.values["gap"] = 0

    def eigenvalues_by_band(self, band=0):
        efermi = self.values["calculation"]["efermi"]
        eigens = np.array(self.values["calculation"]["eband_eigenvalues"])
        return eigens[:, band, 0] - efermi

    def show_eigenvalues_by_band(self, bands=None):
        if bands is None:
            bands = [0]
        kpts = self.values["kpoints"]["list"]
        col_name = {"K-points": kpts}
        for band in bands:
            eigen = self.eigenvalues_by_band(band)
            if self.spin:
                eigens = np.reshape(eigen, [int(len(eigen) / 2), 2])
                name1 = "band" + str(band) + "up"
                name2 = "band" + str(band) + "down"
                col_name[name1] = eigens[:, 0]
                col_name[name2] = eigens[:, 1]
            else:
                name = "band" + str(band)
                col_name[name] = eigen
        df = pd.DataFrame(col_name)
        print(df)

    def export_incar(self, filename=None, print_incar=True):
        """export incar"""
        contents = []
        for key in self.values["incar"]:
            content = key + " = " + str(self.values["incar"][key])
            content += "\n"
            contents.append(str(content))
        if filename is not None:
            with open(filename, "w") as f:
                f.writelines(contents)
        elif print_incar:
            print(contents)
        self.incar = contents

    def export_kpoints(self, filename=None):
        """export kpoints"""
        contents = ["KPOINTS\n"]
        contents += str(len(self.values["kpoints"]["list"])) + "\n"
        contents += ["Cartesian\n"]
        for kpt, wt in zip(self.values["kpoints"]["list"], self.values["kpoints"]["weights"]):
            content = f"{kpt[0]:10.4f} {kpt[1]:10.4f} {kpt[2]:10.4f} {wt[0]:10.4f}"
            if filename is None:
                print(content)
            else:
                content += "\n"
                contents.append(str(content))
        if filename is not None:
            with open(filename, "w") as f:
                f.writelines(contents)

    def export_poscar(self, filename):
        """
        export poscar

        Args:
            filename: string
        Returns:
            a POSCAR file
        """

        comp = self.values["composition"]
        self.values["name_array"]
        latt = self.values["finalpos"]["basis"]
        pos = self.values["finalpos"]["positions"]

        with open(filename, "w") as f:
            string = ""
            for key in comp:
                string += key
                string += str(comp[key])
            string += "\n"
            f.write(string)
            f.write("1.0\n")
            f.write(f"{latt[0][0]:12.6f} {latt[0][1]:12.6f} {latt[0][2]:12.6f}\n")
            f.write(f"{latt[1][0]:12.6f} {latt[1][1]:12.6f} {latt[1][2]:12.6f}\n")
            f.write(f"{latt[2][0]:12.6f} {latt[2][1]:12.6f} {latt[2][2]:12.6f}\n")
            for key in comp:
                f.write(f"{key:4s}")
            f.write("\n")
            for key in comp:
                f.write(f"{comp[key]:4d}")
            f.write("\n")
            f.write("Direct\n")
            for coor in pos:
                f.write(f"{coor[0]:12.6f} {coor[1]:12.6f} {coor[2]:12.6f}\n")

    def parse_bandpath(self):
        kpts = self.values["kpoints"]["list"]
        rec_basis = np.array(self.values["finalpos"]["rec_basis"])

        def inline(kpt, path):
            if len(path) < 2:
                return True
            else:
                v1 = np.array(kpt) - np.array(path[-1])
                v2 = np.array(path[-1]) - np.array(path[-2])
                v1_norm = np.linalg.norm(v1)
                v2_norm = np.linalg.norm(v2)
                if v1_norm < 1e-3:
                    return False
                else:
                    cos = np.dot(v1, v2) / v1_norm / v2_norm
                    if abs(cos - 1) < 1e-2:
                        return True
                    else:
                        # print(kpt, path[-2], path[-1], angle, np.dot(v1, v2)/v1_norm/v2_norm)
                        return False

        paths = []
        path = []
        for i, kpt in enumerate(kpts):
            if inline(kpt, path):
                path.append(kpt)
                if i == len(kpts) - 1:
                    paths.append(path)
            else:
                paths.append(path)
                path = []
                path.append(kpt)

        band_points = []
        pointer = 0
        for i, path in enumerate(paths):
            path = np.array(path)
            dist = np.linalg.norm(np.dot(path[0, :] - path[-1, :], rec_basis))
            x = np.linspace(pointer, pointer + dist, len(path[:, 0]))
            band_paths = x if i == 0 else np.hstack((band_paths, x))
            pointer += dist
            band_points.append(pointer)

        self.values["band_paths"] = band_paths
        self.values["band_points"] = band_points

    def plot_band(
        self,
        filename=None,
        styles="normal",
        ylim=None,
        plim=None,
        saveBands=False,
        dpi=300,
    ):
        """
        plot the bandstructure

        Args:
            filename: string
            styles: string (`normal` or `projected`)
            ylim: list, the range of energy values on the y-axis, e.g. [-5, 3]
            p_max: float (the ratio of color plot in the `projected` mode)

        Returns:
            A figure with band structure
        """
        if plim is None:
            plim = [0.0, 0.5]
        if ylim is None:
            ylim = [-20, 3]
        self.parse_bandpath()
        efermi = self.values["calculation"]["efermi"]
        eigens = np.array(self.values["calculation"]["eband_eigenvalues"])
        paths = self.values["band_paths"]
        band_pts = self.values["band_points"]
        proj = np.array(self.values["calculation"]["projected"])  # [N_kpts, N_band, Ions, 9]
        cm = plt.cm.get_cmap("RdYlBu")
        nkpt, nband, nocc = np.shape(eigens)
        for i in range(nband):
            band = eigens[:, i, 0] - efermi
            if np.all(band < ylim[0]) or np.all(band > ylim[1]):
                continue
            p = np.empty([len(paths)])
            for kpt, _ in enumerate(paths):
                p[kpt] = np.sum(proj[kpt, i, :, :])
            if len(band) / len(paths) == 2:
                plt.plot(paths, band[: len(paths)], c="black", lw=1.0)
                plt.plot(paths, band[len(paths) :], c="red", lw=1.0)
            else:
                plt.plot(paths, band, c="black", lw=1.0)
            if styles == "projected":
                p[p < plim[0]] = plim[0]
                p[p > plim[1]] = plim[1]
                plt.scatter(paths, band, c=p, vmin=plim[0], vmax=plim[1], cmap=cm, s=10)
                if saveBands:
                    np.savetxt("band%04d.dat" % i, np.transpose([band, p]))
            else:
                if saveBands:
                    np.savetxt("band%04d.dat" % i, band)

        for pt in band_pts:
            plt.axvline(x=pt, ls="-", color="k", alpha=0.5)

        if styles == "projected":
            cbar = plt.colorbar(
                orientation="horizontal",
                ticks=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            )
            cbar.set_label("ratio of projected DOS")  # , fontsize=12)
        plt.ylabel("Energy (eV)")
        plt.ylim(ylim)
        plt.xlim([0, paths[-1]])
        plt.xticks([])
        if filename is None:
            plt.show()
        else:
            plt.savefig(filename, dpi=dpi)
            plt.close()

    def get_dos(self, rows, style="t"):
        mydos = []
        labels = []
        a_array = self.values["name_array"]
        N_atom = len(a_array)
        tdos = np.array(self.values["calculation"]["tdos"])
        pdos = np.array(self.values["calculation"]["pdos"])
        a, b, c, d = np.shape(pdos)
        pdos = np.reshape(pdos, [b, a, c, d])
        if style == "t":
            for spin in tdos:
                mydos.append(spin[rows, 1])
                labels.append("total")

        elif style in ["s", "p", "d"]:
            for spin in pdos:
                spd = spin[0, :, :]
                for i in range(1, N_atom):
                    spd += spin[i, :, :]

                if style == "s":
                    mydos.append(spd[rows, 1])
                elif style == "p":
                    mydos.append(spd[rows, 2] + spd[rows, 3] + spd[rows, 4])
                else:
                    mydos.append(spd[rows, 5] + spd[rows, 6] + spd[rows, 7] + spd[rows, 8] + spd[rows, 9])
                labels.append(style)

        elif style[0] == "a":
            if style[1].isdigit():
                ids = style[1].split("-")
                start, end = int(ids[0]), int(ids[1])
                ids = range(start, end + 1)
            else:
                ele = style[1:]
                ids = []
                for i in range(N_atom):
                    if ele == a_array[i]:
                        ids.append(i)
            for spin in pdos:
                spd = spin[ids[0], :, :]
                for i in ids[1:]:
                    spd += spin[i, :, :]
                mydos.append(
                    spd[rows, 1]
                    + spd[rows, 2]
                    + spd[rows, 3]
                    + spd[rows, 4]
                    + spd[rows, 5]
                    + spd[rows, 6]
                    + spd[rows, 7]
                    + spd[rows, 8]
                    + spd[rows, 9]
                )
                labels.append(style[1:])

        if len(labels) == 2:
            labels[0] += "-up"
            labels[1] += "-down"
            mydos[1] *= -1
        return mydos, labels

    def plot_dos(self, filename=None, smear=None, styles="t", xlim=None, dpi=300):
        """
        plot the DOS

        Args:
            filename: string
            styles: string (`t` or `s` or `t+spd`)
            xlim: list, the range of energy values on the x-axis, e.g. [-5, 3]
            smear: float (the width of smearing, defult: None)

        Returns:
            A figure with band structure
        """
        if xlim is None:
            xlim = [-3, 3]
        efermi = self.values["calculation"]["efermi"]
        tdos = np.array(self.values["calculation"]["tdos"][0])
        tdos[:, 0] -= efermi
        e = tdos[:, 0]
        rows = (e > xlim[0]) & (e < xlim[1])
        e = e[rows]
        plt_obj = {}
        for option in styles.split("+"):
            option = ["s", "p", "d"] if option == "spd" else [option]
            for style in option:
                mydos, labels = self.get_dos(rows, style)
                for data, label in zip(mydos, labels):
                    plt_obj[label] = data

        fig, ax = plt.subplots()
        lines1 = []
        lines2 = []
        labels1 = []
        labels2 = []
        for label in plt_obj:
            e = np.reshape(e, [len(e), 1])
            data = np.reshape(plt_obj[label], [len(e), 1])
            if smear is not None:
                data = np.hstack((e, data))
                data = smear_data(data, smear)
                data = data[:, 1]
            if label.find("down") > 0:
                lines2 += ax.plot(e, data)
                labels2.append(label)
            else:
                lines1 += ax.plot(e, data)
                labels1.append(label)
        leg1 = ax.legend(lines1, list(labels1), fancybox=True, loc="upper right")
        leg1.get_frame().set_alpha(0.5)
        if len(lines2) > 0:
            from matplotlib.legend import Legend

            leg2 = Legend(
                ax,
                lines2,
                list(labels2),
                fancybox=True,
                loc="lower right",
            )
            ax.add_artist(leg2)
            leg2.get_frame().set_alpha(0.5)

        plt.xlabel("Energy (eV)")
        plt.ylabel("DOS")
        plt.xlim(xlim)
        if filename is None:
            plt.show()
        else:
            plt.savefig(filename, dpi=dpi)
            plt.close()



"""
This module handles reading and write crystal files.
"""

import importlib.resources

import numpy as np
from monty.serialization import loadfn
from pymatgen.core.bonds import CovalentBond
from pymatgen.core.structure import Molecule, Structure

from pyxtal.constants import logo
from pyxtal.lattice import Lattice
from pyxtal.molecule import Orientation, compare_mol_connectivity, pyxtal_molecule
from pyxtal.msg import ReadSeedError
from pyxtal.symmetry import Group
from pyxtal.util import get_symmetrized_pmg
from pyxtal.wyckoff_site import atom_site, mol_site

with importlib.resources.as_file(importlib.resources.files("pyxtal") / "database" / "bonds.json") as path:
    bonds = loadfn(path)


def in_merged_coords(wp, pt, pts, cell):
    """
    Whether or not the pt in within the pts
    """
    (c, s) = pt
    for pt0 in pts:
        (c0, s0) = pt0
        if s == s0 and wp.are_equivalent_pts(c, c0, cell):
            # print(c, c0, 'equivalent')
            return True
    return False


def get_cif_str_for_pyxtal(struc, header: str = "", sym_num=None, style: str = "mp"):
    """Get the cif string for a given structure. The default setting for
    _atom_site follows the materials project cif

    TODO make this a method of the pyxtal class

    Args:
        struc: pyxtal structure object
        header: additional information
        sym_num: the number of symmetry operations, None means writing all symops
        style: `icsd` or `mp` (used in pymatgen)
    """
    if struc.molecular:
        sites = struc.mol_sites
        molecule = True
        # special = struc.has_special_site()
        # print('===============================================', struc)
    else:
        sites = struc.atom_sites
        molecule = False

    if sym_num is None:
        l_type = struc.group.lattice_type
        number = struc.group.number
        G1 = struc.group[0]
        symbol = struc.group.symbol if G1.is_standard_setting() else sites[0].wp.get_hm_symbol()

    else:  # P1 symmetry
        l_type = "triclinic"
        symbol = "P1"
        number = 1
        G1 = Group(1).Wyckoff_positions[0]

    lines = logo
    lines += "data_" + header + "\n"
    if hasattr(struc, "energy"):
        eng = struc.energy / sum(struc.numMols) if struc.molecular else struc.energy / sum(struc.numIons)
        lines += f"#Energy: {eng} eV/cell\n"

    lines += f"\n_symmetry_space_group_name_H-M '{symbol:s}'\n"
    lines += f"_symmetry_Int_Tables_number      {number:>15d}\n"
    lines += f"_symmetry_cell_setting           {l_type:>15s}\n"

    a, b, c, alpha, beta, gamma = struc.lattice.get_para(degree=True)
    lines += f"_cell_length_a        {a:12.6f}\n"
    lines += f"_cell_length_b        {b:12.6f}\n"
    lines += f"_cell_length_c        {c:12.6f}\n"
    lines += f"_cell_angle_alpha     {alpha:12.6f}\n"
    lines += f"_cell_angle_beta      {beta:12.6f}\n"
    lines += f"_cell_angle_gamma     {gamma:12.6f}\n"
    lines += f"_cell_volume          {struc.lattice.volume:12.6f}\n"
    # if struc.molecular:
    #    lines += '_cell_formula_units_Z     {:d}\n'.format(sum(struc.numMols))
    # else:
    #    lines += '_cell_formula_units_Z     {:d}\n'.format(sum(struc.numIons))

    lines += "\nloop_\n"
    lines += " _symmetry_equiv_pos_site_id\n"
    lines += " _symmetry_equiv_pos_as_xyz\n"

    for i, op in enumerate(G1):
        lines += f"{i + 1:d} '{op.as_xyz_str():s}'\n"

    lines += "\nloop_\n"
    lines += " _atom_site_label\n"
    lines += " _atom_site_type_symbol\n"
    lines += " _atom_site_symmetry_multiplicity\n"
    if style == "icsd":
        lines += " _atom_site_Wyckoff_symbol\n"
    lines += " _atom_site_fract_x\n"
    lines += " _atom_site_fract_y\n"
    lines += " _atom_site_fract_z\n"
    lines += " _atom_site_occupancy\n"

    for site in sites:
        mul = site.wp.multiplicity
        letter = site.wp.letter
        if molecule:
            if sym_num is None:
                coord0s, specie0s = site._get_coords_and_species(first=True)
                if site.wp.index > 0:
                    # print("#Check if the mul is consistent!", site.wp.index)
                    muls = []
                    coords = []
                    species = []
                    merges = []

                    for coord, specie in zip(coord0s, specie0s):
                        _, wp, _ = G1.merge(coord, struc.lattice.matrix, 0.05)
                        if len(wp) > mul:
                            if not in_merged_coords(G1, [coord, specie], merges, struc.lattice.matrix):
                                # print("General Position", specie, coord)
                                coords.append(coord)
                                species.append(specie)
                                muls.append(len(wp))
                                merges.append((coord, specie))
                        else:
                            # print("Special Position", specie, coord)
                            coords.append(coord)
                            species.append(specie)
                            muls.append(mul)

                else:
                    coords, species = coord0s, specie0s
                    muls = [mul] * len(coords)
            else:
                coords = None
                species = []
                for id in range(sym_num):
                    mol = site.get_mol_object(id)
                    tmp = mol.cart_coords.dot(site.lattice.inv_matrix)
                    coords = tmp if coords is None else np.append(coords, tmp, axis=0)
                    species.extend([s.value for s in mol.species])
                muls = [mul] * len(coords)
                # coords, species = site._get_coords_and_species(ids=sym_num)
        else:
            coords, species, muls = [site.position], [site.specie], [mul]

        for specie, coord, mul in zip(species, coords, muls):
            lines += f"{specie:6s} {specie:6s} {mul:3d} "
            if style != "mp":
                lines += f"{letter:s} "
            lines += "{:12.6f}{:12.6f}{:12.6f} 1\n".format(*coord)
    lines += "#END\n\n"

    return lines


def write_cif(struc, filename=None, header="", permission="w", sym_num=None, style="mp"):
    """
    Export the structure in cif format
    The default setting for _atom_site follows the materials project cif

    Args:
        struc: pyxtal structure object
        filename: path of the structure file
        header: additional information
        permission: write(`w`) or append(`a+`) to the given file
        sym_num: the number of symmetry operations, None means writing all symops
        style: `icsd` or `mp` (used in pymatgen)

    """
    lines = get_cif_str_for_pyxtal(struc, header=header, sym_num=sym_num, style=style)

    if filename is None:
        return lines
    else:
        with open(filename, permission) as f:
            f.write(lines)
        return None


def read_cif(filename):
    """
    read the cif, mainly for pyxtal cif output
    Be cautious in using it to read other cif files

    Args:
        filename: path of the structure file

    Return:
        pyxtal structure
    """
    species = []
    coords = []
    with open(filename) as f:
        lines = f.readlines()
        for i, line in enumerate(lines):
            if line.startswith("_symmetry_Int_Tables_number"):
                sg = int(line.split()[-1])
            elif line.startswith("_cell_length_a"):
                a = float(lines[i].split()[-1])
                b = float(lines[i + 1].split()[-1])
                c = float(lines[i + 2].split()[-1])
                alpha = float(lines[i + 3].split()[-1])
                beta = float(lines[i + 4].split()[-1])
                gamma = float(lines[i + 5].split()[-1])
            elif line.startswith("_symmetry_cell_setting"):
                lat_type = line.split()[-1]
            elif line.startswith("_symmetry_space_group_name_H-M "):
                symbol = line.split()[-1]
                diag = eval(symbol) in ["Pn", "P21/n", "C2/n"]

            elif line.find("_atom_site") >= 0:
                s = i
                while True:
                    s += 1
                    if lines[s].find("_atom_site") >= 0:
                        pass
                    elif len(lines[s].split()) <= 3:
                        break
                    else:
                        tmp = lines[s].split()
                        pos = [float(tmp[-4]), float(tmp[-3]), float(tmp[-2])]
                        species.append(tmp[0])
                        coords.append(pos)
                break

    wp0 = Group(sg)[0]
    lattice = Lattice.from_para(a, b, c, alpha, beta, gamma, lat_type)
    sites = []
    for specie, coord in zip(species, coords):
        pt, wp, _ = wp0.merge(coord, lattice.matrix, tol=0.1)
        sites.append(atom_site(wp, pt, specie, diag))
    return lattice, sites


class structure_from_ext:
    def __init__(self, struc, ref_mols, tol=0.2, ignore_HH=False, add_H=False, hn=None):
        """
        extract the mol_site information from the give cif file
        and reference molecule

        Args:
            struc: cif/poscar file or a Pymatgen Structure object
            ref_mols: a list of reference molecule (xyz file or Pyxtal molecule)
            tol: scale factor for covalent bond distance
            ignore_HH: whether or not ignore short H-H in checking molecule
            add_H: whether or not add the H atoms
        """

        for i, ref_mol in enumerate(ref_mols):
            if isinstance(ref_mol, str):
                ref_mols[i] = pyxtal_molecule(ref_mol, fix=True)
            elif isinstance(ref_mol, pyxtal_molecule):
                continue
            else:
                print(type(ref_mol))
                raise NameError(f"reference molecule of type {type(ref_mol)} cannot be defined")

        if isinstance(struc, str):
            pmg_struc = Structure.from_file(struc)
        elif isinstance(struc, Structure):
            pmg_struc = struc
        else:
            print(type(struc))
            raise NameError("input structure cannot be intepretted")

        # reset the hydrogen position
        if add_H:
            pmg_struc.remove_species("H")

        self.ref_mols = ref_mols
        self.tol = tol
        self.add_H = add_H

        sym_struc, number = get_symmetrized_pmg(pmg_struc, hn=hn)
        group = Group(number) if hn is None else Group(hn, use_hall=True)

        self.group = group
        self.wyc = group[0]

        molecules = search_molecules_in_crystal(sym_struc, self.tol, ignore_HH=ignore_HH)

        self.pmg_struc = sym_struc
        matrix = sym_struc.lattice.matrix
        ltype = group.lattice_type
        self.lattice = Lattice.from_matrix(matrix, ltype=ltype)
        self.resort(molecules)
        if len(self.ids) == 0:
            raise RuntimeError("Cannot extract molecules")

    def resort(self, molecules):
        from pyxtal.operations import apply_ops, find_ids

        # filter out the molecular generators
        inv_lat = self.pmg_struc.lattice.inv_matrix
        new_lat = self.lattice.matrix
        positions = np.zeros([len(molecules), 3])
        for i in range(len(molecules)):
            positions[i] = np.dot(molecules[i].cart_coords.mean(axis=0), inv_lat)

        wps = []
        ids = []  # id for the generator
        visited_ids = []
        for id, pos in enumerate(positions):
            if id not in visited_ids:
                centers = apply_ops(pos, self.wyc)
                tmp_ids = find_ids(centers, positions)
                visited_ids.extend(tmp_ids)
                # print(id, pos, tmp_ids, len(self.wyc), len(molecules[id]))
                if len(tmp_ids) == len(self.wyc):
                    # general position
                    # if len(molecules[id])==1: print("groups", tmp_ids, '\n', centers)
                    wps.append(self.wyc)
                    ids.append(id)
                else:  # special sites
                    for id0 in tmp_ids:
                        p0 = positions[id0]
                        p1, wp, _ = self.wyc.merge(p0, new_lat, 0.1)
                        diff = p1 - p0
                        diff -= np.rint(diff)
                        if np.abs(diff).sum() < 1e-2:  # sort position by mapping
                            wps.append(wp)
                            ids.append(id0)  # find the right ids
                            # print("add special", wp.index, id0)
                            break
        # print("===============================================================", self.wps)

        # add position and molecule, print("ids", ids, mults)
        len(ids)
        self.numMols = [0] * len(self.ref_mols)
        self.positions = []
        self.wps = []
        self.p_mols = []
        self.ids = []
        ids_done = []

        # search for the matched molecules
        for j, mol2_ref in enumerate(self.ref_mols):
            mol2 = mol2_ref.copy()
            if self.add_H:
                mol2.mol.remove_species("H")

            for i, id in enumerate(ids):
                mol1 = molecules[id]
                # print("++++++++++++++++++++++++++", id, ids, len(mol2.mol), len(mol1))
                # print(mol2.mol.to('xyz'))
                if id not in ids_done and len(mol2.mol) == len(mol1):
                    p_mol = mol2_ref.copy()  # create p_mol
                    match, mapping = compare_mol_connectivity(mol2.mol, mol1)
                    if match:
                        if len(mol1) > 1:
                            # rearrange the order
                            order = [mapping[at] for at in range(len(mol1))]
                            xyz = mol1.cart_coords[order]
                            # add hydrogen positions here
                            if self.add_H:
                                # print(mol2.smile)
                                xyz = self.add_Hydrogens(mol2.smile, xyz)
                            # print(xyz)
                            frac = np.dot(xyz, inv_lat)
                            xyz = np.dot(frac, new_lat)
                            center = p_mol.get_center(xyz)
                            p_mol.reset_positions(xyz - center)
                            position = np.dot(center, np.linalg.inv(new_lat))
                        else:
                            xyz = mol1.cart_coords[0]
                            position = np.dot(xyz, inv_lat)
                        position -= np.floor(position)

                        self.positions.append(position)
                        self.p_mols.append(p_mol)
                        self.ids.append(j)
                        ids_done.append(id)

                        self.wps.append(wps[i])
                        self.numMols[j] += len(wps[i])
                        # print("================================================ADDDDDD", id, len(mol1))

        # check if some molecules cannot be matched
        if len(ids_done) < len(ids):
            # print("==========================================================Nonmatched molecules", ids_done, ids)
            for id in ids:
                if id not in ids_done:
                    msg = "This molecule cannot be matched to the reference\n"
                    msg += "Molecules extracted from the structure\n"
                    msg += molecules[id].to(fmt="xyz") + "\n"
                    msg += "Reference molecule from smiles or xyz\n"
                    msg += mol2.mol.to(fmt="xyz")
                    raise ReadSeedError(msg)

    def add_Hydrogens(self, smile, xyz):
        """
        add hydrogen for pymtagen molecule
        """
        from rdkit import Chem
        from rdkit.Chem import AllChem
        from rdkit.Geometry import Point3D

        # print("SSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSSS", smile)
        m1 = Chem.MolFromSmiles(smile)
        m2 = Chem.AddHs(m1)
        if len(smile) > 100:
            AllChem.EmbedMolecule(m2, randomSeed=3)
        else:
            AllChem.EmbedMolecule(m2, randomSeed=0xF00D)
        m2 = Chem.RemoveHs(m2)
        conf = m2.GetConformer(0)

        for i in range(conf.GetNumAtoms()):
            x, y, z = xyz[i]
            conf.SetAtomPosition(i, Point3D(x, y, z))
        # Chem.MolToPDBFile(m2, 'test.pdb')
        # conf = m2.GetConformer(0); print(conf.GetPositions())

        m1 = Chem.AddHs(m1)
        if len(smile) > 100:
            AllChem.EmbedMolecule(m1, randomSeed=3)
        else:
            AllChem.EmbedMolecule(m1, randomSeed=0xF00D)

        # print(m1.GetNumAtoms(), m2.GetNumAtoms())
        AllChem.UFFOptimizeMolecule(m1)
        try:
            m3 = AllChem.ConstrainedEmbed(m1, m2)
        except ValueError as e:
            raise ReadSeedError(str(e))

        conf = m3.GetConformer(0)  # ; print(conf.GetPositions())

        return conf.GetPositions()

    def make_mol_sites(self):
        """
        generate the molecular wyckoff sites
        """
        ori = Orientation(np.eye(3))
        sites = []
        for id, mol, pos, wp in zip(self.ids, self.p_mols, self.positions, self.wps):
            # print(id, mol.smile, wp.multiplicity)
            site = mol_site(mol, pos, ori, wp, self.lattice)
            site.type = id
            # print(pos)
            # print(self.lattice.matrix)
            # print([a.value for a in site.molecule.mol.species])
            # print(site.molecule.mol.cart_coords)
            # print(site._get_coords_and_species(absolute=True)[0][:10])
            sites.append(site)
        return sites

    def align(self):
        """
        compute the orientation wrt the reference molecule
        """
        try:
            from openbabel import openbabel, pybel
        except:
            import openbabel
            import pybel

        m1 = pybel.readstring("xyz", self.ref_mol.to("xyz"))
        m2 = pybel.readstring("xyz", self.molecule.to("xyz"))
        aligner = openbabel.OBAlign(True, False)
        aligner.SetRefMol(m1.OBMol)
        aligner.SetTargetMol(m2.OBMol)
        aligner.Align()
        print("RMSD: ", aligner.GetRMSD())
        rot = np.zeros([3, 3])
        for i in range(3):
            for j in range(3):
                rot[i, j] = aligner.GetRotMatrix().Get(i, j)
        coord2 = self.molecule.cart_coords
        coord2 -= np.mean(coord2, axis=0)
        coord3 = rot.dot(coord2.T).T + np.mean(self.ref_mol.cart_coords, axis=0)
        self.mol_aligned = Molecule(self.ref_mol.atomic_numbers, coord3)
        self.ori = Orientation(rot)

    def show(self, overlay=True):
        from pyxtal.viz import display_molecules

        if overlay:
            return display_molecules([self.ref_mol, self.mol_aligned])
        else:
            return display_molecules([self.ref_mol, self.molecule])


def search_molecules_in_crystal(struc, tol=0.2, once=False, ignore_HH=True, max_bond_length=None):
    """
    Function to perform to find the molecule in a Pymatgen structure

    Args:
        struc: Pymatgen Structure
        tol: tolerance value to check the connectivity
        once: search only one molecule or all molecules
        ignore_HH: whether or not ignore the short H-H in checking molecule
        max_bond_length: sets maximum bond length if bond length is missing in bond length database

    Returns:
        molecules: list of pymatgen molecules
        positions: list of center positions
    """

    def check_one_layer(struc, sites0, visited):
        new_members = []
        for site0 in sites0:
            sites_add, visited = check_one_site(struc, site0, visited)
            new_members.extend(sites_add)
        return new_members, visited

    def check_one_site(struc, site0, visited, rmax=2.8):
        neigh_sites = struc.get_neighbors(site0, rmax)
        ids = [m.index for m in visited]
        sites_add = []
        ids_add = []
        pbc = isinstance(struc, Structure)

        for site1 in neigh_sites:
            if site1.index not in ids + ids_add:
                try:
                    if CovalentBond.is_bonded(site0, site1, tol):
                        if pbc:
                            (d, image) = site0.distance_and_image(site1)
                        else:
                            d = site0.distance(site1)
                        val1, val2 = site1.specie.value, site0.specie.value
                        key = f"{val1:s}-{val2:s}"

                        # sometime the H-H short distance is not avoidable
                        if key == "H-H":
                            if not ignore_HH:
                                if pbc:
                                    site1.frac_coords += image
                                sites_add.append(site1)
                                ids_add.append(site1.index)
                        else:
                            if d < bonds.get(key,max_bond_length):
                                if pbc:
                                    site1.frac_coords += image
                                sites_add.append(site1)
                                ids_add.append(site1.index)
                except ValueError:
                    # QZ: use our own bond distance lib
                    if pbc:
                        (d, image) = site0.distance_and_image(site1)
                    else:
                        d = site0.distance(site1)

                    val1, val2 = site1.specie.value, site0.specie.value
                    key = f"{val1:s}-{val2:s}"
                    if d < bonds[key]:
                        if pbc:
                            site1.frac_coords += image
                        sites_add.append(site1)
                        ids_add.append(site1.index)

        if len(sites_add) > 0:
            visited.extend(sites_add)

        return sites_add, visited

    molecules = []
    visited_ids = []

    for id, site in enumerate(struc.sites):
        if id not in visited_ids:
            first_site = site
            visited = [first_site]
            first_site.index = id
            n_iter, max_iter = 0, len(struc) - len(visited_ids)
            while n_iter < max_iter:
                if n_iter == 0:
                    new_sites, visited = check_one_site(struc, first_site, visited)
                else:
                    new_sites, visited = check_one_layer(struc, new_sites, visited)
                n_iter += 1
                if len(new_sites) == 0:
                    break

            coords = [s.coords for s in visited]
            coords = np.array(coords)
            numbers = [s.specie.number for s in visited]
            molecules.append(Molecule(numbers, coords))
            visited_ids.extend([s.index for s in visited])
            # print(molecules[-1].to(fmt='xyz')); import sys; sys.exit()
        if once and len(molecules) == 1:
            break

    return molecules


if __name__ == "__main__":
    from pyxtal.database.collection import Collection

    pmg = Structure.from_file("pyxtal/database/cifs/resorcinol.cif")
    mols = search_molecules_in_crystal(pmg, tol=0.2, once=False)
    print(len(mols))

    pmg = Collection("molecules")["xxv"]
    mols = search_molecules_in_crystal(pmg, tol=0.2, once=False)
    print(len(mols))



"""
crystal plane class
"""

import itertools
from math import gcd

import numpy as np


def has_reduction(hkl):
    h, k, l = hkl
    gcf = gcd(h, gcd(k, l))
    if gcf > 1:
        # like [2, 2, 0]
        return True
    elif hkl[np.nonzero(np.array(hkl))[0][0]] < 0:
        # like [-2, 0, 0]
        return True
    return False


def reduced_hkl(hkl):
    h, k, l = hkl
    gcf = gcd(h, gcd(k, l))
    if gcf > 1:
        return [int(h / gcf), int(k / gcf), int(l / gcf)], gcf
    else:
        return [h, k, l], 1


def structure_factor(pos, hkl, total=True):
    coords = np.dot(pos, hkl)
    F = np.exp(-2 * np.pi * (1j) * coords)
    if total:
        return F.sum()
    else:
        return F


def get_dspacing(inv_matrix, hkl):
    return 1 / np.linalg.norm(inv_matrix.dot(np.array(hkl)))


class planes:
    """
    This is a database class to process crystal data

    Args:
        db_name: *.db format from ase database
        d_min: the minimum layer spacing
        cp_factor: skip non-close packed plane
    """

    def __init__(self, extent=6, d_min=1.5, cp_factor=0.5):
        self.d_min = d_min
        self.extent = extent
        self.cp_factor = cp_factor
        self.set_planes()
        # self.set_xtal()

    def set_xtal(self, xtal):
        self.xtal = xtal
        self.atoms = xtal.to_ase(center_only=True)
        self.cell_reciprocal = xtal.lattice.inv_matrix

    def set_planes(self):
        planes = list(itertools.product(range(-self.extent, self.extent + 1), repeat=3))
        planes = [hkl for hkl in planes if hkl != (0, 0, 0) and not has_reduction(hkl)]
        self.planes = planes

    def get_cp_factor(self, hkl):
        hkl, hkl_factor = reduced_hkl(hkl)
        hkl = np.array(hkl)
        dspacing = get_dspacing(self.cell_reciprocal, hkl)
        cp_factor = 0
        for n in range(1, 10):
            if dspacing / n > self.d_min:
                _factor = np.abs(self.get_structure_factor(n * hkl)) / len(self.atoms)
                if _factor > cp_factor:
                    cp_factor = _factor
            else:
                break
        return cp_factor

    def search_close_packing_planes(self, N_max=10):
        """
        Search for the close-packed molecular plane for a given crystal

        Args:
            N_max: maximum number of multiples
        """
        cp_planes = []
        for hkl in self.planes:
            dspacing = get_dspacing(self.cell_reciprocal, hkl)
            for n in range(1, N_max):
                if dspacing / n > self.d_min:
                    hkl1 = n * np.array(hkl)
                    F = self.get_structure_factor(hkl1)
                    if np.abs(F) >= len(self.atoms) * self.cp_factor:
                        # Scan the plane with high density
                        plane = self.get_separation(hkl1)
                        if plane is not None:
                            cp_planes.append(plane)
                        break
                else:
                    # early termination for very short ranged hkl
                    break
        if len(cp_planes) > 0:
            cp_planes = sorted(cp_planes, key=lambda x: -x[-1][0])
        return cp_planes

    def get_structure_factor(self, hkl):
        return structure_factor(self.atoms.get_scaled_positions(), hkl)

    def get_separation(self, hkl):
        """
        Compute the separation for the given hkl plane

        Args:
            - hkl: three indices
        """
        hkl_reduced, hkl_factor = reduced_hkl(hkl)
        d_spacing = get_dspacing(self.cell_reciprocal, hkl_reduced)

        # Go over each molecule to compute the molecular slab
        slabs = []
        for mol_site in self.xtal.mol_sites:
            N_atoms = len(mol_site.numbers)
            center_frac = mol_site.position  # fractional
            xyz, species = mol_site._get_coords_and_species(unitcell=True)
            for id_mol, op in enumerate(mol_site.wp.ops):
                start, end = id_mol * N_atoms, (id_mol + 1) * N_atoms
                coords = xyz[start:end, :]  # frac
                center_frac = op.operate(mol_site.position)
                center_frac -= np.floor(center_frac)
                coords -= center_frac  # place the center to (0, 0, 0)
                center_hkl = np.dot(center_frac, hkl_reduced)
                center_hkl -= np.floor(center_hkl)
                coords_hkl = np.dot(coords, hkl_reduced)

                lower, upper = (
                    center_hkl + coords_hkl.min(),
                    center_hkl + coords_hkl.max(),
                )
                slabs.append([center_hkl, lower, upper])
        # if np.abs(hkl-np.array([0, 8, 0])).sum()==0:
        groups = self.group_slabs(slabs, 0.5 / hkl_factor)
        groups = self.group_slabs(groups, 0.5 / hkl_factor)
        # print(hkl); print(groups)
        separations = self.find_unique_separations(groups, d_spacing)
        return (hkl, d_spacing / abs(hkl_factor), separations)

    def group_slabs(self, slabs, tol):
        groups = []
        for slab in slabs:
            new = True
            center, lower, upper = slab
            for group in groups:
                # Center is within the group
                if (
                    group[1] <= center <= group[2]
                    or lower >= group[1]
                    and upper <= group[2]
                    or lower <= group[1]
                    and upper >= group[2]
                ):
                    new = False
                # to include
                elif lower - 1 <= group[1] and upper - 1 >= group[2] or lower - 1 >= group[1] and upper - 1 <= group[2]:
                    center, lower, upper = center - 1, lower - 1, upper - 1
                    new = False
                else:
                    dist = center - group[0]
                    shift = np.rint(dist)
                    if abs(dist - shift) < tol:
                        new = False
                        lower -= shift
                        upper -= shift

                if not new:
                    # Update group
                    if lower < group[1]:
                        group[1] = lower
                    if upper > group[2]:
                        group[2] = upper
                    center = (group[1] + group[2]) / 2
                    break
            if new:
                groups.append([center, lower, upper])
        return sorted(groups)

    def find_unique_separations(self, groups, d):
        groups.append(groups[0])
        groups[-1] = [g + 1 for g in groups[-1]]
        separations = []
        for i in range(len(groups) - 1):
            separations.append(d * (groups[i + 1][1] - groups[i][2]))
        separations = np.unique(-1 * np.array(separations).round(decimals=3))
        return -np.sort(separations)

    def gather(self, planes, tol=-0.1):
        output = []
        for _plane in planes:
            (hkl, d, separations) = _plane
            if separations[0] <= d:
                if separations[0] > tol:
                    for separation in separations:
                        if separation > tol:
                            p = plane(hkl, self.cell_reciprocal, separation)
                            print(p)
                            output.append(_plane)
                else:
                    break
        return output


class plane:
    """
    This simplest possible plane object
    """

    def __init__(self, hkl, cell_reciprocal, separation=None):
        self.indices = hkl
        self.cell_reciprocal = cell_reciprocal
        self.simple_indices, self.factor = reduced_hkl(hkl)
        self.d_spacing = get_dspacing(self.cell_reciprocal, self.indices)
        self.separation = separation

    def __str__(self):
        s = "({:2d} {:2d} {:2d})".format(*self.indices)
        s += f" Spacing: {self.d_spacing:6.3f}"
        s += f" Separation: {self.separation:6.3f}"
        return s


if __name__ == "__main__":
    import sys

    from pyxtal.db import database

    try:
        from ccdc import io
        from ccdc.particle import SlipPlanes

        csd = io.EntryReader("CSD")
    except:
        csd = None
        print("Cannot import ccdc to check")

    db = database("pyxtal/database/mech.db")

    if len(sys.argv) == 1:
        codes = ["ANLINB02", "ADIPAC", "CHEXDC"]
        for code in [c for c in db.codes if c not in codes]:
            # for code in codes:
            print("\n", code)
            p = planes()
            p.set_xtal(db.get_pyxtal(code))
            cp_planes = p.search_close_packing_planes()
            ps = p.gather(cp_planes, 0)

            # Crossvalidation with csd_python
            if csd is not None:
                splanes = SlipPlanes(csd.crystal(code))
                splanes = [
                    splane
                    for splane in splanes
                    if splane.repeat_distance > p.d_min and p.get_cp_factor(list(splane.orientation.hkl)) > p.cp_factor
                ]
                if len(splanes) > len(ps):
                    for splane in splanes:
                        print(
                            code,
                            splane.orientation,
                            round(splane.repeat_distance, 3),
                            round(splane.slab_separation, 3),
                        )
    else:
        data = [
            ("UCECAG01", [3, 0, -2]),
            ("PIDGOZ", [1, 0, 2]),
        ]
        p = planes(cp_factor=0.2)
        for code, hkl in data:
            print(hkl, tuple(hkl) in p.planes)
            p.set_xtal(db.get_pyxtal(code))
            print(code, hkl, p.get_separation(hkl))



from warnings import warn

from pyxtal.constants import pyxtal_verbosity


def printx(text, priority=1):
    """
    Custom printing function based on verbosity.

    Args:
        text: string to be passed to print
        priority: the importance of printing the message
            0: Critical; must be printed
            1: Warning; unexpected error but program functioning
            2: Info; useful but not necessary print out
            3: Debug; detailed information for debugging

    Returns:
        Nothing
    """
    if priority <= 1:
        warn(text)
        return
    else:
        if priority <= pyxtal_verbosity:
            print(text)


class Error(Exception):
    """Base class for exceptions in this module."""


class ConformerError(Error):
    """Exception raised for errors in the Compabality.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class Symm_CompatibilityError(Error):
    """Exception raised for errors in the Compabality.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class Comp_CompatibilityError(Error):
    """Exception raised for errors in the Compabality.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class ReadSeedError(Error):
    """Exception raised for errors in the Compabality.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class VolumeError(Error):
    """Exception raised for errors in the Compabality.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class CSDError(Error):
    """Exception raised for errors in the Compabality.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message


class AtomTypeError(Error):
    """Exception raised for errors in the Compabality.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, message):
        self.message = message



import os
import time

import numpy as np
from ase import Atoms
from ase.calculators.vasp import Vasp
from ase.io import read

from pyxtal import pyxtal
from pyxtal.util import good_lattice

"""
A script to perform multistages vasp calculation
"""


class VASP:
    """
    This is a calculator to perform structure optimization in GULP
    At the moment, only inorganic crystal is considered

    Args:

    struc: structure object generated by Pyxtal
    ff: path of forcefield lib
    opt: `conv`, `conp`, `single`
    """

    def __init__(self, struc, path="tmp", cmd="mpirun -np 16 vasp_std"):
        if isinstance(struc, pyxtal):
            struc = struc.to_ase()

        if not isinstance(struc, Atoms):
            raise NotImplementedError("only support ASE atoms object")

        self.structure = struc
        self.folder = path
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)
        self.pstress = 0.0
        self.energy = None
        self.energy_per_atom = None
        self.stress = None
        self.forces = None
        self.gap = None
        self.cputime = 0
        self.error = True
        self.cmd = cmd

    def set_vasp(self, level=0, pstress=0.0000, setup=None):
        self.pstress = pstress
        default0 = {
            "xc": "pbe",
            "npar": 8,
            "kgamma": True,
            "lcharg": False,
            "lwave": False,
            "ibrion": 2,
            "pstress": pstress * 10,
            "setups": setup,
        }
        if level == 0:
            default1 = {
                "prec": "low",
                "algo": "normal",
                "kspacing": 0.4,
                "isif": 4,
                "ediff": 1e-2,
                "nsw": 10,
                "potim": 0.02,
            }
        elif level == 1:
            default1 = {
                "prec": "normal",
                "algo": "normal",
                "kspacing": 0.3,
                "isif": 3,
                "ediff": 1e-3,
                "nsw": 25,
                "potim": 0.05,
            }
        elif level == 2:
            default1 = {
                "prec": "accurate",
                "kspacing": 0.2,
                "isif": 3,
                "ediff": 1e-3,
                "nsw": 50,
                "potim": 0.1,
            }
        elif level == 3:
            default1 = {
                "prec": "accurate",
                "encut": 600,
                "kspacing": 0.15,
                "isif": 3,
                "ediff": 1e-4,
                "nsw": 50,
            }
        elif level == 4:
            default1 = {
                "prec": "accurate",
                "encut": 600,
                "kspacing": 0.15,
                "isif": 3,
                "ediff": 1e-4,
                "nsw": 0,
            }

        dict_vasp = dict(default0, **default1)
        return Vasp(**dict_vasp)

    def read_OUTCAR(self, path="OUTCAR"):
        """read time and ncores info from OUTCAR"""
        time = 0
        ncore = 0
        for line in open(path):
            if line.rfind("running on  ") > -1:
                ncore = int(line.split()[2])
            elif line.rfind("Elapsed time ") > -1:
                time = float(line.split(":")[-1])
        self.cputime = time
        self.ncore = ncore

    def read_OSZICAR(self, path="OSZICAR"):
        """read the enthalpy from OSZICAR"""
        energy = 100000
        for line in open(path):
            if line.rfind(" F= ") > -1:
                energy = float(line.split()[2])
        self.energy = energy  # this is actually enthalpy

    def read_bandgap(self, path="vasprun.xml"):
        from pyxtal.interface.vasprun import vasprun

        myrun = vasprun(path)
        self.gap = myrun.values["gap"]

    def run(self, setup=None, pstress=0, level=0, clean=True, read_gap=False, walltime=None):
        if walltime is not None:
            os.environ["VASP_COMMAND"] = "timeout " + max_time + " " + self.cmd
        else:
            os.environ["VASP_COMMAND"] = self.cmd
        cwd = os.getcwd()
        setups = self.set_vasp(level, pstress, setup)
        self.structure.set_calculator(setups)
        os.chdir(self.folder)
        try:
            self.structure.get_potential_energy()
            self.error = False
            self.read_OSZICAR()
        except RuntimeError:
            # VASP is not full done
            self.read_OSZICAR()
            if self.energy < 10000:
                self.error = False
        except (IndexError, ValueError, UnboundLocalError):
            print("Error in parsing vasp output or VASP calc is wrong")
            os.system("cp OUTCAR Error-OUTCAR")

        if not self.error:
            try:
                self.forces = self.structure.get_forces()
            except:
                self.forces = np.zeros([len(self.structure), 3])
            self.energy_per_atom = self.energy / len(self.structure)
            self.read_OUTCAR()
            if read_gap:
                self.read_bandgap()
        if clean:
            self.clean()

        os.chdir(cwd)

    def clean(self):
        os.remove("POSCAR")
        os.remove("POTCAR")
        os.remove("INCAR")
        os.remove("OUTCAR")
        if os.path.exists("OSZICAR"):
            os.remove("OSZICAR")

    def to_pymatgen(self):
        from pymatgen.core.structure import Structure

        return Structure(self.lattice.matrix, self.sites, self.frac_coords)

    def to_pyxtal(self):
        struc = pyxtal()
        struc.from_seed(self.structure)
        return struc


def single_optimize(
    struc,
    level,
    pstress,
    setup,
    path,
    clean,
    cmd="mpirun -np 16 vasp_std",
    walltime="30m",
):
    """
    single optmization

    Args:
        struc: pyxtal structure
        level: vasp calc level
        pstress: external pressure
        setup: vasp setup
        path: calculation directory

    Returns:
        the structure, energy and time costs
    """
    calc = VASP(struc, path, cmd=cmd)
    calc.run(setup, pstress, level, clean=clean)
    if calc.error:
        return None, None, 0, True
    else:
        try:
            struc = calc.to_pyxtal()
            struc.optimize_lattice()
            return struc, calc.energy_per_atom, calc.cputime, calc.error
        except:
            return None, None, 0, True


def single_point(struc, setup=None, path=None, clean=True):
    """
    single optmization

    Args:
        struc: pyxtal structure
        level: vasp calc level
        pstress: external pressure
        setup: vasp setup
        path: calculation directory

    Returns:
        the energy and forces
    """
    calc = VASP(struc, path)
    calc.run(setup, level=4, clean=clean)
    return calc.energy, calc.forces, calc.error


def optimize(
    struc,
    path,
    levels=None,
    pstress=0,
    setup=None,
    clean=True,
    cmd="mpirun -np 16 vasp_std",
    walltime="30m",
):
    """
    multi optimization

    Args:
        struc: pyxtal structure
        path: calculation directory
        levels: list of vasp calc levels
        pstress: external pressure
        setup: vasp setup

    Returns:
        list of structures, energies and time costs
    """

    if levels is None:
        levels = [0, 2, 3]
    time_total = 0
    for _i, level in enumerate(levels):
        struc, eng, time, error = single_optimize(struc, level, pstress, setup, path, clean, cmd, walltime)

        time_total += time
        # print(eng, time, time_total, '++++++++++++++++++++++++++++++')
        if error or not good_lattice(struc):
            return None, None, 0, True
    return struc, eng, time_total, error


def VASP_relax(struc, opt_cell=False, step=100, kspacing=0.25, pstress=0, folder="tmp"):
    if not os.path.exists(folder):
        os.makedirs(folder)
    cwd = os.getcwd()
    os.chdir(folder)

    isif = 3 if opt_cell else 2

    calc = Vasp(
        xc="PBE",
        prec="Accurate",
        npar=8,
        kgamma=True,
        lcharg=False,
        lwave=False,
        ibrion=2,
        pstress=pstress,
        kspacing=kspacing,
        nsw=step,
        isif=isif,
        potim=0.05,
        ediff=1e-3,
    )
    struc.set_calculator(calc)
    energy = struc.get_potential_energy()
    struc = read("CONTCAR", format="vasp")
    os.chdir(cwd)
    return struc, energy


if __name__ == "__main__":
    while True:
        struc = pyxtal()
        struc.from_random(3, 19, ["C"], [4])
        if struc.valid:
            break

    # set up the commands
    os.system("source /share/intel/mkl/bin/mklvars.sh intel64")
    cmd = "mpirun -n 4 /share/apps/bin/vasp544-2019u2/vasp_std"

    calc = VASP(struc, path="tmp", cmd=cmd)
    calc.run()
    print("Energy:", calc.energy)
    print("Forces", calc.forces)

    struc, eng, time, _ = optimize(struc, path="tmp", levels=[0, 1, 2], cmd=cmd, walltime="30s")
    print(struc)
    print("Energy:", eng)
    print("Time:", time)

    calc = VASP(struc, path="tmp", cmd=cmd)
    calc.run(level=4, read_gap=True)
    print("Energy:", calc.energy)
    print("Gap:", calc.gap)



from distutils.core import setup
#from os import path
import setuptools  # noqa

#this_directory = path.abspath(path.dirname(__file__))
#with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
#    long_description = f.read()

with open("README.md") as fh:
    long_description = fh.read()

setup(
    name="pyxtal",
    version="1.0.1",
    author="Scott Fredericks, Kevin Parrish, Qiang Zhu",
    author_email="alecfans@gmail.com",
    description="Python code for generation of crystal structures based on symmetry constraints.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MaterSim/PyXtal",
    packages=[
        "pyxtal",
        "pyxtal.database",
        "pyxtal.interface",
        "pyxtal.optimize",
        "pyxtal.potentials",
        "pyxtal.database.cifs",
    ],
    package_data={
        "pyxtal.database": ["*.csv", "*.json", "*.db"],
        "pyxtal.database.cifs": ["*.cif", "*.vasp"],
        "pyxtal.potentials": ["*"],
    },
    scripts=[
        "scripts/pyxtal_main.py",
        "scripts/pyxtal_symmetry.py",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "spglib>=1.10.4",
        "pymatgen>=2024.3.1",
        "pandas>=2.0.2",
        "networkx>=2.3",
        "ase>=3.23.0",
        "scipy>=1.7.3",
        "numpy>=1.26,<2",  # prevent the use of numpy2
        "importlib_metadata>=1.4",
        "typing-extensions>=4.12",
    ],
    extras_require={
        "visualization": ["py3Dmol>=0.8.0"],
        "descriptor": ["pyshtools>=4.10.3"],
        "molecules": ["openbabel", "pybel"],
        "test": ["wheel", "pytest", "coverage", "pytest-cov", "monty>=2024.2.26"],
    },
    python_requires=">=3.9",
    license="MIT",
)



# python -m unittest pyxtal/test_all.py
import importlib.util
import os
import unittest

from pyxtal import pyxtal
from pyxtal.lattice import Lattice
from pyxtal.symmetry import Group, Wyckoff_position
from pyxtal.util import generate_wp_lib


def resource_filename(package_name, resource_path):
    package_path = importlib.util.find_spec(package_name).submodule_search_locations[0]
    return os.path.join(package_path, resource_path)


cif_path = resource_filename("pyxtal", "database/cifs/")
l01 = Lattice.from_matrix([[4.08, 0, 0], [0, 9.13, 0], [0, 0, 5.50]])
l02 = Lattice.from_para(4.08, 9.13, 5.50, 90, 90, 90)
wp1 = Wyckoff_position.from_group_and_index(36, 0)
wp2 = Wyckoff_position.from_group_and_letter(36, "4a")


class TestGroup(unittest.TestCase):
    def test_generate_wp_lib(self):
        wps = generate_wp_lib([227, 228], composition=[1, 2])
        assert len(wps) == 18
        wps = generate_wp_lib([227, 228], composition=[1, 1, 3])
        assert len(wps) == 9

    def test_list_wyckoff_combinations(self):
        g = Group(64)
        a1, _, _ = g.list_wyckoff_combinations([4, 2])
        assert len(a1) == 0
        a2, _, _ = g.list_wyckoff_combinations([4, 8], quick=False)
        assert len(a2) == 8

    def test_print_group_and_dof(self):
        for d in [(1, 6), (15, 4), (60, 3), (143, 2), (208, 1)]:
            (sg, dof_ref) = d
            g = Group(sg)
            dof = g.get_lattice_dof()
            assert dof == dof_ref

    def test_get_spg_symmetry_object(self):
        spg_list = [14, 36, 62, 99, 143, 160, 182, 191, 225, 230]
        ans = [32, 18, 36, 21, 16, 19, 24, 48, 62, 62]
        for spg, num in zip(spg_list, ans):
            g = Group(spg)
            ss = g.get_spg_symmetry_object()
            matrix = ss.to_matrix_representation_spg()
            assert num == sum(sum(matrix))

    def test_short_path(self):
        g = Group(217)
        path = g.short_path_to_general_wp(7)
        assert path[-1][2] == 145

    def test_spg_symmetry(self):
        N_polar, N_centro, N_chiral = 0, 0, 0
        for sg in range(1, 231):
            g = Group(sg, quick=True)
            _pg, polar, centro, chiral = g.point_group, g.polar, g.inversion, g.chiral
            if polar:
                N_polar += 1
            if centro:
                N_centro += 1
            if chiral:
                N_chiral += 1
        assert N_polar == 68
        assert N_centro == 92
        assert N_chiral == 65

    def test_ferroelectric(self):
        pairs = [(4, 1), (187, 4), (222, 5)]
        for pair in pairs:
            (sg, N) = pair
            assert len(Group(sg, quick=True).get_ferroelectric_groups()) == N

    def test_check_compatible(self):
        assert Group(225).check_compatible([64, 28, 24]) == (True, True)
        assert Group(227).check_compatible([8]) == (True, False)
        assert Group(227).check_compatible([4]) == (False, False)
        assert Group(19).check_compatible([6]) == (False, False)

    def test_search_supergroup_paths(self):
        paths = Group(59, quick=True).search_supergroup_paths(139, 2)
        assert paths == [[71, 139], [129, 139], [137, 139]]

    def test_get_splitters(self):
        s = pyxtal()
        s.from_seed(cif_path + "3-G139.cif")
        g = Group(225)
        solutions = g.get_splitters_from_structure(s, "t")
        assert len(solutions) == 3

    def test_add_k_transitions(self):
        paras = [
            (189, 26, 6),
            (11, 6, 2),
            (62, 33, 3),
            (63, 33, 5),
        ]

        for p in paras:
            (g, h, n_path) = p
            gr = Group(g, quick=True)
            path = gr.search_subgroup_paths(h)[0]
            solutions = gr.add_k_transitions(path)
            assert len(solutions) == n_path

    def test_to_subgroup(self):
        s = pyxtal(molecular=True)
        s.from_seed(cif_path + "benzene.cif", ["benzene"])
        c = s.to_subgroup()
        assert c.valid

    def test_get_wyckoff_position_from_xyz(self):
        g = Group(5)
        pos = [
            ([0.0, 0.0, 0.0], "2a"),
            ([0.5, 0.5, 0.5], None),
            ([0.1, 0.0, 0.1], "4c"),
            ([0.5, 0.0, 0.0], None),
            ([0.0, 0.1, 0.0], "2a"),
            ([0.0, 0.5, 0.5], "2b"),
            ([0.1, 0.2, 0.3], "4c"),
        ]
        for d in pos:
            (p0, wp0) = d
            wp = g.get_wyckoff_position_from_xyz(p0)
            if wp is None:
                assert wp0 is None
            else:
                assert wp.get_label() == wp0

if __name__ == "__main__":
    unittest.main()
